levelbar block.warning-battery-offset {
  background-color: @error_color;
  border-color: @error_color;
}

levelbar block.low-battery-offset {
  background-color: @warning_color;
  border-color: @warning_color;
}

levelbar block.high-battery-offset {
  background-color: @success_color;
  border-color: @success_color;
}
